import os, glob, pathlib, shutil, subprocess
from pathlib import Path
from typing import List
from dflow.python import (
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    upload_packages
)
try:
    from monty.serialization import loadfn
    from apex.utils import return_prop_list
    from apex.core.common_equi import (make_equi, post_equi)
    from .utils import recursive_search
except:
    pass

upload_packages.append(__file__)


class RelaxMake(OP):
    """
    OP class for making calculation tasks
    """

    def __init__(self):
        pass

    @classmethod
    def get_input_sign(cls):
        return OPIOSign({
            'input': Artifact(Path),
            'param': dict
        })

    @classmethod
    def get_output_sign(cls):
        return OPIOSign({
            'output': Artifact(Path),
            'njobs': int,
            'task_names': List[str],
            'task_paths': Artifact(List[Path])
        })

    @OP.exec_sign_check
    def execute(
            self,
            op_in: OPIO,
    ) -> OPIO:
        from apex.core.common_equi import make_equi

        cwd = os.getcwd()
        os.chdir(op_in["input"])
        work_d = os.getcwd()
        param_argv = op_in["param"]
        structures = param_argv["structures"]
        inter_parameter = param_argv["interaction"]
        parameter = param_argv["relaxation"]

        make_equi(structures, inter_parameter, parameter)

        conf_dirs = []
        for conf in structures:
            conf_dirs.extend(glob.glob(conf))
        conf_dirs.sort()

        task_list = []
        task_list_str = []
        for ii in conf_dirs:
            conf_dir_global = os.path.join(work_d, ii)
            task_list.append(os.path.join(conf_dir_global, 'relaxation/relax_task'))
            task_list_str.append(os.path.join(ii, 'relaxation'))

        all_jobs = task_list
        njobs = len(all_jobs)
        jobs = []
        for job in all_jobs:
            jobs.append(pathlib.Path(job))

        os.chdir(cwd)
        op_out = OPIO({
            "output": op_in["input"],
            "task_names": task_list_str,
            "njobs": njobs,
            "task_paths": jobs
        })
        return op_out


class RelaxPost(OP):
    """
    OP class for analyzing calculation results
    """

    def __init__(self):
        pass

    @classmethod
    def get_input_sign(cls):
        return OPIOSign({
            'input_post': Artifact(Path, sub_path=False),
            'input_all': Artifact(Path),
            'param': dict
        })

    @classmethod
    def get_output_sign(cls):
        return OPIOSign({
            'retrieve_path': Artifact(List[Path]),
            'output_all': Artifact(Path)
        })

    @OP.exec_sign_check
    def execute(self, op_in: OPIO) -> OPIO:
        from apex.core.common_equi import post_equi
        cwd = os.getcwd()
        param_argv = op_in['param']
        inter_param = param_argv["interaction"]
        calculator = inter_param["type"]
        conf_list = param_argv["structures"]
        copy_dir_list_input = [conf.split('/')[0] for conf in conf_list]
        os.chdir(op_in['input_all'])
        copy_dir_list = []
        for ii in copy_dir_list_input:
            copy_dir_list.extend(glob.glob(ii))

        # find path of finished tasks
        os.chdir(op_in['input_post'])
        src_path = recursive_search(copy_dir_list)
        if not src_path:
            raise RuntimeError(f'Fail to find input work path after slices!')

        os.chdir(op_in['input_all'])
        if calculator in ['vasp', 'abacus']:
            shutil.copytree(op_in['input_post'], './', dirs_exist_ok=True)
            post_equi(conf_list, inter_param)
        else:
            # src_path = str(input_post) + str(local_path)
            shutil.copytree(src_path, './', dirs_exist_ok=True)
            post_equi(conf_list, inter_param)
            conf_dirs = []
            for conf in conf_list:
                conf_dirs.extend(glob.glob(conf))
            conf_dirs.sort()
            for ii in conf_dirs:
                cmd = 'rm *.pb'
                os.chdir(ii)
                subprocess.call(cmd, shell=True)
                os.chdir(op_in['input_all'])
                os.chdir(os.path.join(ii, 'relaxation/relax_task'))
                subprocess.call(cmd, shell=True)
                os.chdir(op_in['input_all'])

        os.chdir(cwd)
        for ii in copy_dir_list:
            src_path = str(op_in['input_all']) + f'/{ii}'
            shutil.copytree(src_path, f'./{ii}', dirs_exist_ok=True)
        post_path = [Path(ii) for ii in copy_dir_list]

        op_out = OPIO({
            'retrieve_path': post_path,
            'output_all': op_in['input_all']
        })
        return op_out



#!/usr/bin/python3
import os

import dpdata
import numpy as np
from pymatgen.core.structure import Structure

import apex.core.calculator.lib.abacus_scf as abacus_scf
from dflow.python import upload_packages
upload_packages.append(__file__)

A2BOHR = 1.8897261254578281
MASS_DICT = {
    "H": 1.0079,
    "He": 4.0026,
    "Li": 6.941,
    "Be": 9.0122,
    "B": 10.811,
    "C": 12.0107,
    "N": 14.0067,
    "O": 15.9994,
    "F": 18.9984,
    "Ne": 20.1797,
    "Na": 22.9897,
    "Mg": 24.305,
    "Al": 26.9815,
    "Si": 28.0855,
    "P": 30.9738,
    "S": 32.065,
    "Cl": 35.453,
    "K": 39.0983,
    "Ar": 39.948,
    "Ca": 40.078,
    "Sc": 44.9559,
    "Ti": 47.867,
    "V": 50.9415,
    "Cr": 51.9961,
    "Mn": 54.938,
    "Fe": 55.845,
    "Ni": 58.6934,
    "Co": 58.9332,
    "Cu": 63.546,
    "Zn": 65.39,
    "Ga": 69.723,
    "Ge": 72.64,
    "As": 74.9216,
    "Se": 78.96,
    "Br": 79.904,
    "Kr": 83.8,
    "Rb": 85.4678,
    "Sr": 87.62,
    "Y": 88.9059,
    "Zr": 91.224,
    "Nb": 92.9064,
    "Mo": 95.94,
    "Tc": 98,
    "Ru": 101.07,
    "Rh": 102.9055,
    "Pd": 106.42,
    "Ag": 107.8682,
    "Cd": 112.411,
    "In": 114.818,
    "Sn": 118.71,
    "Sb": 121.76,
    "I": 126.9045,
    "Te": 127.6,
    "Xe": 131.293,
    "Cs": 132.9055,
    "Ba": 137.327,
    "La": 138.9055,
    "Ce": 140.116,
    "Pr": 140.9077,
    "Nd": 144.24,
    "Pm": 145,
    "Sm": 150.36,
    "Eu": 151.964,
    "Gd": 157.25,
    "Tb": 158.9253,
    "Dy": 162.5,
    "Ho": 164.9303,
    "Er": 167.259,
    "Tm": 168.9342,
    "Yb": 173.04,
    "Lu": 174.967,
    "Hf": 178.49,
    "Ta": 180.9479,
    "W": 183.84,
    "Re": 186.207,
    "Os": 190.23,
    "Ir": 192.217,
    "Pt": 195.078,
    "Au": 196.9665,
    "Hg": 200.59,
    "Tl": 204.3833,
    "Pb": 207.2,
    "Bi": 208.9804,
    "Po": 209,
    "At": 210,
    "Rn": 222,
    "Fr": 223,
    "Ra": 226,
    "Ac": 227,
    "Pa": 231.0359,
    "Th": 232.0381,
    "Np": 237,
    "U": 238.0289,
    "Am": 243,
    "Pu": 244,
    "Cm": 247,
    "Bk": 247,
    "Cf": 251,
    "Es": 252,
    "Fm": 257,
    "Md": 258,
    "No": 259,
    "Rf": 261,
    "Lr": 262,
    "Db": 262,
    "Bh": 264,
    "Sg": 266,
    "Mt": 268,
    "Rg": 272,
    "Hs": 277,
    "H": 1.0079,
    "He": 4.0026,
    "Li": 6.941,
    "Be": 9.0122,
    "B": 10.811,
    "C": 12.0107,
    "N": 14.0067,
    "O": 15.9994,
    "F": 18.9984,
    "Ne": 20.1797,
    "Na": 22.9897,
    "Mg": 24.305,
    "Al": 26.9815,
    "Si": 28.0855,
    "P": 30.9738,
    "S": 32.065,
    "Cl": 35.453,
    "K": 39.0983,
    "Ar": 39.948,
    "Ca": 40.078,
    "Sc": 44.9559,
    "Ti": 47.867,
    "V": 50.9415,
    "Cr": 51.9961,
    "Mn": 54.938,
    "Fe": 55.845,
    "Ni": 58.6934,
    "Co": 58.9332,
    "Cu": 63.546,
    "Zn": 65.39,
    "Ga": 69.723,
    "Ge": 72.64,
    "As": 74.9216,
    "Se": 78.96,
    "Br": 79.904,
    "Kr": 83.8,
    "Rb": 85.4678,
    "Sr": 87.62,
    "Y": 88.9059,
    "Zr": 91.224,
    "Nb": 92.9064,
    "Mo": 95.94,
    "Tc": 98,
    "Ru": 101.07,
    "Rh": 102.9055,
    "Pd": 106.42,
    "Ag": 107.8682,
    "Cd": 112.411,
    "In": 114.818,
    "Sn": 118.71,
    "Sb": 121.76,
    "I": 126.9045,
    "Te": 127.6,
    "Xe": 131.293,
    "Cs": 132.9055,
    "Ba": 137.327,
    "La": 138.9055,
    "Ce": 140.116,
    "Pr": 140.9077,
    "Nd": 144.24,
    "Pm": 145,
    "Sm": 150.36,
    "Eu": 151.964,
    "Gd": 157.25,
    "Tb": 158.9253,
    "Dy": 162.5,
    "Ho": 164.9303,
    "Er": 167.259,
    "Tm": 168.9342,
    "Yb": 173.04,
    "Lu": 174.967,
    "Hf": 178.49,
    "Ta": 180.9479,
    "W": 183.84,
    "Re": 186.207,
    "Os": 190.23,
    "Ir": 192.217,
    "Pt": 195.078,
    "Au": 196.9665,
    "Hg": 200.59,
    "Tl": 204.3833,
    "Pb": 207.2,
    "Bi": 208.9804,
    "Po": 209,
    "At": 210,
    "Rn": 222,
    "Fr": 223,
    "Ra": 226,
    "Ac": 227,
    "Pa": 231.0359,
    "Th": 232.0381,
    "Np": 237,
    "U": 238.0289,
    "Am": 243,
    "Pu": 244,
    "Cm": 247,
    "Bk": 247,
    "Cf": 251,
    "Es": 252,
    "Fm": 257,
    "Md": 258,
    "No": 259,
    "Rf": 261,
    "Lr": 262,
    "Db": 262,
    "Bh": 264,
    "Sg": 266,
    "Mt": 268,
    "Rg": 272,
    "Hs": 277,
}
key_words_list = [
    "ATOMIC_SPECIES",
    "NUMERICAL_ORBITAL",
    "LATTICE_CONSTANT",
    "LATTICE_VECTORS",
    "ATOMIC_POSITIONS",
    "NUMERICAL_DESCRIPTOR",
]


def poscar2stru(poscar, inter_param, stru):
    """
    - poscar:           POSCAR for input
    - inter_param:      dictionary of 'interaction' from param.json
                        some key words for ABACUS are:
                            - atom_masses:  a dictionary of atoms' masses
                            - orb_files:     a dictionary of orbital files
                            - deepks_desc:  a string of deepks descriptor file
    - stru:            output filename, usally is 'STRU'
    """
    stru = dpdata.System(poscar, fmt="vasp/poscar")
    stru_data = stru.data
    atom_mass = []
    pseudo = None
    orb = None
    deepks_desc = None

    if "atom_masses" not in inter_param:
        atom_mass_dict = {
            i: 1.0 if i not in MASS_DICT else MASS_DICT[i]
            for i in stru_data["atom_names"]
        }
    else:
        atom_mass_dict = inter_param["atom_masses"]
    for atom in stru_data["atom_names"]:
        assert atom in atom_mass_dict, (
            "the mass of %s is not defined in interaction:atom_masses" % atom
        )
        atom_mass.append(atom_mass_dict[atom])

    if "potcars" in inter_param:
        pseudo = []
        for atom in stru_data["atom_names"]:
            assert atom in inter_param["potcars"], (
                "the pseudopotential of %s is not defined in interaction:potcars" % atom
            )
            pseudo.append("./pp_orb/" + inter_param["potcars"][atom].split("/")[-1])

    if "orb_files" in inter_param:
        orb = []
        for atom in stru_data["atom_names"]:
            assert atom in inter_param["orb_files"], (
                "orbital file of %s is not defined in interaction:orb_files" % atom
            )
            orb.append("./pp_orb/" + inter_param["orb_files"][atom].split("/")[-1])

    if "deepks_desc" in inter_param:
        deepks_desc = "./pp_orb/%s\n" % inter_param["deepks_desc"]

    stru.to(
        "stru",
        "STRU",
        mass=atom_mass,
        pp_file=pseudo,
        numerical_orbital=orb,
        numerical_descriptor=deepks_desc,
    )


def stru_fix_atom(struf, fix_atom=[True, True, True]):
    """
    ...
    ATOMIC_POSITIONS
    Cartesian               #Cartesian(Unit is LATTICE_CONSTANT)
    Si                      #Name of element
    0.0                     #Magnetic for this element.
    2                       #Number of atoms
    0.00 0.00 0.00 0 0 0    #x,y,z, move_x, move_y, move_z
    0.25 0.25 0.25 0 0 0
    """
    fix_xyz = ["0" if i else "1" for i in fix_atom]
    if os.path.isfile(struf):
        with open(struf) as f1:
            lines = f1.readlines()
        for i in range(len(lines)):
            if "ATOMIC_POSITIONS" in lines[i]:
                break
        i += 1
        flag_read_coord_type = False
        flag_read_atom_number = 2
        flag_atom_number = 0
        while i < len(lines):
            if lines[i].strip() == "":
                pass
            elif lines[i].split()[0] in key_words_list:
                break
            elif not flag_read_coord_type:
                flag_read_coord_type = True
            elif flag_atom_number:
                flag_atom_number -= 1
                x, y, z = lines[i].split()[:3]
                lines[i] = "%s %s %s %s %s %s\n" % tuple([x, y, z] + fix_xyz)
            elif flag_read_coord_type and flag_read_atom_number:
                flag_read_atom_number -= 1
            elif not flag_read_atom_number:
                flag_read_atom_number = 2
                flag_atom_number = int(lines[i].split()[0])
            i += 1
        with open(struf, "w") as f1:
            f1.writelines(lines)
    else:
        raise RuntimeError(
            "Error: Try to modify struc file %s, but can not find it" % struf
        )


def stru_scale(stru_in, stru_out, scale):
    with open(stru_in, "r") as fin:
        lines = fin.readlines()
    for i in range(len(lines)):
        if "LATTICE_CONSTANT" in lines[i]:
            lines[i + 1] = str(float(lines[i + 1].strip()) * scale) + "\n"
            break
    with open(stru_out, "w") as f1:
        f1.writelines(lines)


def write_kpt(kptf, kptlist):
    context = "K_POINTS\n0\nGamma\n"
    for i in kptlist:
        context += str(i) + " "
    with open(kptf, "w") as f1:
        f1.write(context)


def write_input(inputf, inputdict):
    context = "INPUT_PARAMETERS\n"
    for key in inputdict.keys():
        if key[0] in ["_", "#"]:
            continue
        context += key + " " + str(inputdict[key]) + "\n"
    with open(inputf, "w") as f1:
        f1.write(context)


def make_kspacing_kpt(struf, kspacing):
    stru_data = abacus_scf.get_abacus_STRU(struf)
    cell = stru_data["cells"] / abacus_scf.bohr2ang
    volume = abs(cell[0].dot(np.cross(cell[1], cell[2])))
    coef = 2 * np.pi / volume / kspacing
    kpt = [
        max(1, int(np.linalg.norm(np.cross(cell[x], cell[y])) * coef + 1))
        for x, y in [[1, 2], [2, 0], [0, 1]]
    ]
    return kpt


def check_finished(fname):
    with open(fname, "r") as fp:
        return "Total  Time  :" in fp.read()


def final_stru(abacus_path):
    with open(os.path.join(abacus_path, "INPUT")) as f1:
        lines = f1.readlines()
    suffix = "ABACUS"
    calculation = "scf"
    out_stru = False
    for line in lines:
        if "suffix" in line and line.split()[0] == "suffix":
            suffix = line.split()[1]
        elif "calculation" in line and line.split()[0] == "calculation":
            calculation = line.split()[1]
        elif "out_stru" in line and line.split()[0] == "out_stru":
            out_stru = bool(line.split()[1])
    logf = os.path.join(abacus_path, "OUT.%s/running_%s.log" % (suffix, calculation))
    if calculation in ["relax", "cell-relax"]:
        if not out_stru:
            return "OUT.%s/STRU_ION_D" % suffix
        else:
            with open(logf) as f1:
                lines = f1.readlines()
            for i in range(1, len(lines)):
                if lines[-i][36:41] == "istep":
                    max_step = int(lines[-i].split()[-1])
                    break
            return "OUT.%s/STRU_ION%d_D" % (suffix, max_step)
    elif calculation == "md":
        with open(logf) as f1:
            lines = f1.readlines()
        for i in range(1, len(lines)):
            if lines[-i][1:27] == "STEP OF MOLECULAR DYNAMICS":
                max_step = int(lines[-i].split()[-1])
                break
        return "OUT.%s/STRU_MD_%d" % (suffix, max_step)
    elif calculation == "scf":
        return "STRU"
    else:
        print("Unrecognized calculation type in %s/INPUT" % abacus_path)
        return "STRU"


def stru2Structure(struf):
    stru = dpdata.System(struf, fmt="stru")
    stru.to("poscar", "POSCAR.tmp")
    ss = Structure.from_file("POSCAR.tmp")
    os.remove("POSCAR.tmp")
    return ss


def check_stru_fixed(struf, fixed):
    block = {}
    with open(struf) as f1:
        lines = f1.readlines()
    for line in lines:
        if line.strip() == "":
            continue
        elif line.split()[0] in key_words_list:
            key = line.split()[0]
            block[key] = []
        else:
            block[key].append(line)
    i = 3
    while i < len(block["ATOMIC_POSITIONS"]):
        natom = int(block["ATOMIC_POSITIONS"][i])
        for j in range(natom):
            i += 1
            for k in block["ATOMIC_POSITIONS"][i].split()[3:6]:
                if fixed and bool(int(k)):
                    return False
                elif not fixed and not bool(int(k)):
                    return False
        i += 1
    return True


def modify_stru_path(strucf, tpath):
    if tpath[-1] != "/":
        tpath += "/"
    with open(strucf) as f1:
        lines = f1.readlines()
    for i, line in enumerate(lines):
        if "ATOMIC_SPECIES" in line and line.split()[0] == "ATOMIC_SPECIES":
            file_numb = 2
        elif (
            "NUMERICAL_ORBITAL" in line and line.split()[0] == "NUMERICAL_ORBITAL"
        ) or (
            "NUMERICAL_DESCRIPTOR" in line and line.split()[0] == "NUMERICAL_DESCRIPTOR"
        ):
            file_numb = 0
        else:
            continue

        for j in range(i + 1, len(lines)):
            if lines[j].strip() in key_words_list:
                break
            elif lines[j].strip() == "":
                continue
            ppfile = tpath + os.path.split(lines[j].split()[file_numb])[1]
            tmp_line = ""
            for k in range(file_numb):
                tmp_line += lines[j].split()[k] + " "
            lines[j] = tmp_line + ppfile + "\n"

    with open(strucf, "w") as f1:
        f1.writelines(lines)




#!/usr/bin/env python3

import os
import random
import subprocess as sp
import sys

import dpdata
from dpdata.periodic_table import Element
from packaging.version import Version

import apex.core.lib.util as util
from dflow.python import upload_packages
upload_packages.append(__file__)


def cvt_lammps_conf(fin, fout, type_map, ofmt="lammps/data"):
    """
    Format convert from fin to fout, specify the output format by ofmt
    Imcomplete situation
    """
    supp_ofmt = ["lammps/dump", "lammps/data", "vasp/poscar"]
    supp_exts = ["dump", "lmp", "poscar/POSCAR"]

    if "dump" in fout:
        ofmt = "lammps/dump"
    elif "lmp" in fout:
        ofmt = "lammps/data"
    elif "poscar" in fout or "POSCAR" in fout:
        ofmt = "vasp/poscar"
    if not ofmt in supp_ofmt:
        raise RuntimeError(
            "output format " + ofmt + " is not supported. use one of " + str(supp_ofmt)
        )

    if "lmp" in fout:
        d_poscar = dpdata.System(fin, fmt="vasp/poscar", type_map=type_map)
        d_poscar.to_lammps_lmp(fout, frame_idx=0)
    elif "poscar" in fout or "POSCAR" in fout:
        d_dump = dpdata.System(fin, fmt="lammps/dump", type_map=type_map)
        d_dump.to_vasp_poscar(fout, frame_idx=-1)


def apply_type_map(conf_file, deepmd_type_map, ptypes):
    """
    apply type map.
    conf_file:          conf file converted from POSCAR
    deepmd_type_map:    deepmd atom type map
    ptypes:             atom types defined in POSCAR
    """
    natoms = _get_conf_natom(conf_file)
    ntypes = len(deepmd_type_map)
    with open(conf_file, "r") as fp:
        lines = fp.read().split("\n")
    # with open(conf_file+'.bk', 'w') as fp:
    #     fp.write("\n".join(lines))
    new_lines = lines
    # revise ntypes
    idx_ntypes = -1
    for idx, ii in enumerate(lines):
        if "atom types" in ii:
            idx_ntypes = idx
    if idx_ntypes == -1:
        raise RuntimeError("cannot find the entry 'atom types' in ", conf_file)
    words = lines[idx_ntypes].split()
    words[0] = str(ntypes)
    new_lines[idx_ntypes] = " ".join(words)
    # find number of atoms
    idx_atom_entry = -1
    for idx, ii in enumerate(lines):
        if "Atoms" in ii:
            idx_atom_entry = idx
    if idx_atom_entry == -1:
        raise RuntimeError("cannot find the entry 'Atoms' in ", conf_file)
    # revise atom type
    for idx in range(idx_atom_entry + 2, idx_atom_entry + 2 + natoms):
        ii = lines[idx]
        words = ii.split()
        assert len(words) >= 5
        old_id = int(words[1])
        new_id = deepmd_type_map.index(ptypes[old_id - 1]) + 1
        words[1] = str(new_id)
        ii = " ".join(words)
        new_lines[idx] = ii
    with open(conf_file, "w") as fp:
        fp.write("\n".join(new_lines))


def _get_ntype(conf):
    with open(conf, "r") as fp:
        lines = fp.read().split("\n")
    for ii in lines:
        if "atom types" in ii:
            return int(ii.split()[0])
    raise RuntimeError("cannot find line indicate atom types in ", conf)


def _get_conf_natom(conf):
    with open(conf, "r") as fp:
        lines = fp.read().split("\n")
    for ii in lines:
        if "atoms" in ii:
            return int(ii.split()[0])
    raise RuntimeError("cannot find line indicate atom types in ", conf)


def inter_deepmd(param):
    models = param["model_name"]
    deepmd_version = param["deepmd_version"]
    ret = "pair_style deepmd "
    model_list = ""
    for ii in models:
        model_list += ii + " "
    if Version(deepmd_version) < Version("1"):
        ## DeePMD-kit version == 0.x
        if len(models) > 1:
            ret += "%s 10 model_devi.out\n" % model_list
        else:
            ret += models[0] + "\n"
    else:
        ## DeePMD-kit version >= 1
        if len(models) > 1:
            ret += "%s out_freq 10 out_file model_devi.out\n" % model_list
        else:
            ret += models[0] + "\n"
    ret += "pair_coeff * *\n"
    return ret


def inter_meam(param):
    ret = ""
    line = "pair_style      meam \n"
    line += "pair_coeff      * * %s " % param["model_name"][0]
    for ii in param["param_type"]:
        line += ii + " "
    line += "%s " % param["model_name"][1]
    for ii in param["param_type"]:
        line += ii + " "
    line += "\n"
    ret += line
    return ret


def inter_eam_fs(param):  # 06/08 eam.fs interaction
    ret = ""
    line = "pair_style      eam/fs \n"
    line += "pair_coeff      * * %s " % param["model_name"][0]
    for ii in param["param_type"]:
        line += ii + " "
    line += "\n"
    ret += line
    return ret


def inter_eam_alloy(param):  # 06/08 eam.alloy interaction
    ret = ""
    line = "pair_style      eam/alloy \n"
    line += "pair_coeff      * * %s " % param["model_name"][0]
    for ii in param["param_type"]:
        line += ii + " "
    line += "\n"
    ret += line
    return ret


def element_list(type_map):
    type_map_reverse = {k: v for v, k in type_map.items()}
    type_map_list = []
    tmp_list = list(type_map_reverse.keys())
    tmp_list.sort()
    for ii in tmp_list:
        type_map_list.append(type_map_reverse[ii])
    return type_map_list


def make_lammps_eval(conf, type_map, interaction, param):
    type_map_list = element_list(type_map)

    """
    make lammps input for static calcualtion
    """
    ret = ""
    ret += "clear\n"
    ret += "units 	metal\n"
    ret += "dimension	3\n"
    ret += "boundary	p p p\n"
    ret += "atom_style	atomic\n"
    ret += "box         tilt large\n"
    ret += "read_data   %s\n" % conf
    for ii in range(len(type_map)):
        ret += "mass            %d %.3f\n" % (ii + 1, Element(type_map_list[ii]).mass)
    ret += "neigh_modify    every 1 delay 0 check no\n"
    ret += interaction(param)
    ret += "compute         mype all pe\n"
    ret += "thermo          100\n"
    ret += (
        "thermo_style    custom step pe pxx pyy pzz pxy pxz pyz lx ly lz vol c_mype\n"
    )
    ret += "dump            1 all custom 100 dump.relax id type xs ys zs fx fy fz\n"  # 06/09 give dump.relax
    ret += "run    0\n"
    ret += "variable        N equal count(all)\n"
    ret += "variable        V equal vol\n"
    ret += 'variable        E equal "c_mype"\n'
    ret += "variable        tmplx equal lx\n"
    ret += "variable        tmply equal ly\n"
    ret += "variable        Pxx equal pxx\n"
    ret += "variable        Pyy equal pyy\n"
    ret += "variable        Pzz equal pzz\n"
    ret += "variable        Pxy equal pxy\n"
    ret += "variable        Pxz equal pxz\n"
    ret += "variable        Pyz equal pyz\n"
    ret += "variable        Epa equal ${E}/${N}\n"
    ret += "variable        Vpa equal ${V}/${N}\n"
    ret += "variable        AA equal (${tmplx}*${tmply})\n"
    ret += 'print "All done"\n'
    ret += 'print "Total number of atoms = ${N}"\n'
    ret += 'print "Final energy per atoms = ${Epa}"\n'
    ret += 'print "Final volume per atoms = ${Vpa}"\n'
    ret += 'print "Final Base area = ${AA}"\n'
    ret += 'print "Final Stress (xx yy zz xy xz yz) = ${Pxx} ${Pyy} ${Pzz} ${Pxy} ${Pxz} ${Pyz}"\n'
    return ret


def make_lammps_equi(
    conf,
    type_map,
    interaction,
    param,
    etol=0,
    ftol=1e-10,
    maxiter=5000,
    maxeval=500000,
    change_box=True,
):
    type_map_list = element_list(type_map)

    """
    make lammps input for equilibritation
    """
    ret = ""
    ret += "clear\n"
    ret += "units 	metal\n"
    ret += "dimension	3\n"
    ret += "boundary	p p p\n"
    ret += "atom_style	atomic\n"
    ret += "box         tilt large\n"
    ret += "read_data   %s\n" % conf
    for ii in range(len(type_map)):
        ret += "mass            %d %.3f\n" % (ii + 1, Element(type_map_list[ii]).mass)
    ret += "neigh_modify    every 1 delay 0 check no\n"
    ret += interaction(param)
    ret += "compute         mype all pe\n"
    ret += "thermo          100\n"
    ret += (
        "thermo_style    custom step pe pxx pyy pzz pxy pxz pyz lx ly lz vol c_mype\n"
    )
    ret += "dump            1 all custom 100 dump.relax id type xs ys zs fx fy fz\n"
    ret += "min_style       cg\n"
    if change_box:
        ret += "fix             1 all box/relax iso 0.0 \n"
        ret += "minimize        %e %e %d %d\n" % (etol, ftol, maxiter, maxeval)
        ret += "fix             1 all box/relax aniso 0.0 \n"
        ret += "minimize        %e %e %d %d\n" % (etol, ftol, maxiter, maxeval)
        ret += "fix             1 all box/relax tri 0.0 \n"
    ret += "minimize        %e %e %d %d\n" % (etol, ftol, maxiter, maxeval)
    ret += "variable        N equal count(all)\n"
    ret += "variable        V equal vol\n"
    ret += 'variable        E equal "c_mype"\n'
    ret += "variable        tmplx equal lx\n"
    ret += "variable        tmply equal ly\n"
    ret += "variable        Pxx equal pxx\n"
    ret += "variable        Pyy equal pyy\n"
    ret += "variable        Pzz equal pzz\n"
    ret += "variable        Pxy equal pxy\n"
    ret += "variable        Pxz equal pxz\n"
    ret += "variable        Pyz equal pyz\n"
    ret += "variable        Epa equal ${E}/${N}\n"
    ret += "variable        Vpa equal ${V}/${N}\n"
    ret += "variable        AA equal (${tmplx}*${tmply})\n"
    ret += 'print "All done"\n'
    ret += 'print "Total number of atoms = ${N}"\n'
    ret += 'print "Final energy per atoms = ${Epa}"\n'
    ret += 'print "Final volume per atoms = ${Vpa}"\n'
    ret += 'print "Final Base area = ${AA}"\n'
    ret += 'print "Final Stress (xx yy zz xy xz yz) = ${Pxx} ${Pyy} ${Pzz} ${Pxy} ${Pxz} ${Pyz}"\n'
    return ret


def make_lammps_elastic(
    conf, type_map, interaction, param, etol=0, ftol=1e-10, maxiter=5000, maxeval=500000
):
    type_map_list = element_list(type_map)

    """
    make lammps input for elastic calculation
    """
    ret = ""
    ret += "clear\n"
    ret += "units 	metal\n"
    ret += "dimension	3\n"
    ret += "boundary	p p p\n"
    ret += "atom_style	atomic\n"
    ret += "box         tilt large\n"
    ret += "read_data   %s\n" % conf
    for ii in range(len(type_map)):
        ret += "mass            %d %.3f\n" % (ii + 1, Element(type_map_list[ii]).mass)
    ret += "neigh_modify    every 1 delay 0 check no\n"
    ret += interaction(param)
    ret += "compute         mype all pe\n"
    ret += "thermo          100\n"
    ret += (
        "thermo_style    custom step pe pxx pyy pzz pxy pxz pyz lx ly lz vol c_mype\n"
    )
    ret += "dump            1 all custom 100 dump.relax id type xs ys zs fx fy fz\n"
    ret += "min_style       cg\n"
    ret += "minimize        %e %e %d %d\n" % (etol, ftol, maxiter, maxeval)
    ret += "variable        N equal count(all)\n"
    ret += "variable        V equal vol\n"
    ret += 'variable        E equal "c_mype"\n'
    ret += "variable        Pxx equal pxx\n"
    ret += "variable        Pyy equal pyy\n"
    ret += "variable        Pzz equal pzz\n"
    ret += "variable        Pxy equal pxy\n"
    ret += "variable        Pxz equal pxz\n"
    ret += "variable        Pyz equal pyz\n"
    ret += "variable        Epa equal ${E}/${N}\n"
    ret += "variable        Vpa equal ${V}/${N}\n"
    ret += 'print "All done"\n'
    ret += 'print "Total number of atoms = ${N}"\n'
    ret += 'print "Final energy per atoms = ${Epa}"\n'
    ret += 'print "Final volume per atoms = ${Vpa}"\n'
    ret += 'print "Final Stress (xx yy zz xy xz yz) = ${Pxx} ${Pyy} ${Pzz} ${Pxy} ${Pxz} ${Pyz}"\n'
    return ret


def make_lammps_press_relax(
    conf,
    type_map,
    scale2equi,
    interaction,
    param,
    B0=70,
    bp=0,
    etol=0,
    ftol=1e-10,
    maxiter=5000,
    maxeval=500000,
):
    type_map_list = element_list(type_map)

    """
    make lammps input for relaxation at a certain volume
    scale2equi: the volume scale with respect to equilibrium volume
    """
    ret = ""
    ret += "clear\n"
    ret += "variable        GPa2bar	equal 1e4\n"
    ret += "variable        B0		equal %f\n" % B0
    ret += "variable        bp		equal %f\n" % bp
    ret += "variable	    xx		equal %f\n" % scale2equi
    ret += "variable        yeta	equal 1.5*(${bp}-1)\n"
    ret += (
        "variable        Px0		equal 3*${B0}*(1-${xx})/${xx}^2*exp(${yeta}*(1-${xx}))\n"
    )
    ret += "variable        Px		equal ${Px0}*${GPa2bar}\n"
    ret += "units       metal\n"
    ret += "dimension   3\n"
    ret += "boundary	p p p\n"
    ret += "atom_style	atomic\n"
    ret += "box         tilt large\n"
    ret += "read_data   %s\n" % conf
    for ii in range(len(type_map)):
        ret += "mass            %d %.3f\n" % (ii + 1, Element(type_map_list[ii]).mass)
    ret += "neigh_modify    every 1 delay 0 check no\n"
    ret += interaction(param)
    ret += "compute         mype all pe\n"
    ret += "thermo          100\n"
    ret += (
        "thermo_style    custom step pe pxx pyy pzz pxy pxz pyz lx ly lz vol c_mype\n"
    )
    ret += "dump            1 all custom 100 dump.relax id type xs ys zs fx fy fz\n"
    ret += "min_style       cg\n"
    ret += "fix             1 all box/relax iso ${Px} \n"
    ret += "minimize        %e %e %d %d\n" % (etol, ftol, maxiter, maxeval)
    ret += "fix             1 all box/relax aniso ${Px} \n"
    ret += "minimize        %e %e %d %d\n" % (etol, ftol, maxiter, maxeval)
    ret += "variable        N equal count(all)\n"
    ret += "variable        V equal vol\n"
    ret += 'variable        E equal "c_mype"\n'
    ret += "variable        Pxx equal pxx\n"
    ret += "variable        Pyy equal pyy\n"
    ret += "variable        Pzz equal pzz\n"
    ret += "variable        Pxy equal pxy\n"
    ret += "variable        Pxz equal pxz\n"
    ret += "variable        Pyz equal pyz\n"
    ret += "variable        Epa equal ${E}/${N}\n"
    ret += "variable        Vpa equal ${V}/${N}\n"
    ret += 'print "All done"\n'
    ret += 'print "Total number of atoms  = ${N}"\n'
    ret += 'print "Relax at Press         = ${Px} Bar"\n'
    ret += 'print "Final energy per atoms = ${Epa} eV"\n'
    ret += 'print "Final volume per atoms = ${Vpa} A^3"\n'
    ret += 'print "Final Stress (xx yy zz xy xz yz) = ${Pxx} ${Pyy} ${Pzz} ${Pxy} ${Pxz} ${Pyz}"\n'
    return ret


def make_lammps_phonon(
    conf, masses, interaction, param, etol=0, ftol=1e-10, maxiter=5000, maxeval=500000
):
    """
    make lammps input for elastic calculation
    """
    ret = ""
    ret += "clear\n"
    ret += "units 	metal\n"
    ret += "dimension	3\n"
    ret += "boundary	p p p\n"
    ret += "atom_style	atomic\n"
    ret += "box         tilt large\n"
    ret += "read_data   %s\n" % conf
    ntypes = len(masses)
    for ii in range(ntypes):
        ret += "mass            %d %f\n" % (ii + 1, masses[ii])
    ret += "neigh_modify    every 1 delay 0 check no\n"
    ret += interaction(param)
    return ret


def _get_epa(lines):
    for ii in lines:
        if ("Final energy per atoms" in ii) and (not "print" in ii):
            return float(ii.split("=")[1].split()[0])
    raise RuntimeError(
        'cannot find key "Final energy per atoms" in lines, something wrong'
    )


def _get_vpa(lines):
    for ii in lines:
        if ("Final volume per atoms" in ii) and (not "print" in ii):
            return float(ii.split("=")[1].split()[0])
    raise RuntimeError(
        'cannot find key "Final volume per atoms" in lines, something wrong'
    )


def _get_natoms(lines):
    for ii in lines:
        if ("Total number of atoms" in ii) and (not "print" in ii):
            return int(ii.split("=")[1].split()[0])
    raise RuntimeError(
        'cannot find key "Total number of atoms" in lines, something wrong'
    )


def get_nev(log):
    """
    get natoms, energy_per_atom and volume_per_atom from lammps log
    """
    with open(log, "r") as fp:
        lines = fp.read().split("\n")
    epa = _get_epa(lines)
    vpa = _get_vpa(lines)
    natoms = _get_natoms(lines)
    return natoms, epa, vpa


def get_base_area(log):
    """
    get base area
    """
    with open(log, "r") as fp:
        lines = fp.read().split("\n")
    for ii in lines:
        if ("Final Base area" in ii) and (not "print" in ii):
            return float(ii.split("=")[1].split()[0])


def get_stress(log):
    """
    get stress from lammps log
    """
    with open(log, "r") as fp:
        lines = fp.read().split("\n")
    for ii in lines:
        if ("Final Stress" in ii) and (not "print" in ii):
            vstress = [float(jj) for jj in ii.split("=")[1].split()]
    stress = util.voigt_to_stress(vstress)
    return stress


def poscar_from_last_dump(dump, poscar_out, deepmd_type_map):
    """
    get poscar from the last frame of a lammps MD traj (dump format)
    """
    with open(dump, "r") as fp:
        lines = fp.read().split("\n")
    step_idx = -1
    for idx, ii in enumerate(lines):
        if "ITEM: TIMESTEP" in ii:
            step_idx = idx
    if step_idx == -1:
        raise RuntimeError("cannot find timestep in lammps dump, something wrong")
    with open("tmp_dump", "w") as fp:
        fp.write("\n".join(lines[step_idx:]))
    cvt_lammps_conf("tmp_dump", poscar_out, ofmt="vasp")
    os.remove("tmp_dump")
    with open(poscar_out, "r") as fp:
        lines = fp.read().split("\n")
    types = [deepmd_type_map[int(ii.split("_")[1])] for ii in lines[5].split()]
    lines[5] = " ".join(types)
    with open(poscar_out, "w") as fp:
        lines = fp.write("\n".join(lines))


def check_finished_new(fname, keyword):
    with open(fname, "r") as fp:
        lines = fp.read().split("\n")
    flag = False
    for jj in lines:
        if (keyword in jj) and (not "print" in jj):
            flag = True
    return flag


def check_finished(fname):
    with open(fname, "r") as fp:
        return "Total wall time:" in fp.read()




#!/usr/bin/env python3

import logging
import os
import re
import shutil
from dflow.python import upload_packages
upload_packages.append(__file__)

iter_format = "%s"
task_format = "%s"
log_iter_head = "task type: " + iter_format + " task: " + task_format + " process: "


def make_iter_name(iter_index):
    return "task type:" + (iter_format)


def create_path(path):
    path += "/"
    if os.path.isdir(path):
        dirname = os.path.dirname(path)
        counter = 0
        while True:
            bk_dirname = dirname + ".bk%03d" % counter
            if not os.path.isdir(bk_dirname):
                shutil.move(dirname, bk_dirname)
                break
            counter += 1
    os.makedirs(path)


def replace(file_name, pattern, subst):
    file_handel = open(file_name, "r")
    file_string = file_handel.read()
    file_handel.close()
    file_string = re.sub(pattern, subst, file_string)
    file_handel = open(file_name, "w")
    file_handel.write(file_string)
    file_handel.close()


def copy_file_list(file_list, from_path, to_path):
    for jj in file_list:
        if os.path.isfile(os.path.join(from_path, jj)):
            shutil.copy(os.path.join(from_path, jj), to_path)
        elif os.path.isdir(os.path.join(from_path, jj)):
            shutil.copytree(os.path.join(from_path, jj), os.path.join(to_path, jj))


def cmd_append_log(cmd, log_file):
    ret = cmd
    ret = ret + " 1> " + log_file
    ret = ret + " 2> " + log_file
    return ret


def log_iter(task, ii, jj):
    logging.info((log_iter_head + "%s") % (ii, jj, task))


def repeat_to_length(string_to_expand, length):
    ret = ""
    for ii in range(length):
        ret += string_to_expand
    return ret


def log_task(message):
    header = repeat_to_length(" ", len(log_iter_head % (0, 0)))
    logging.info(header + message)


def record_iter(record, confs, ii, jj):
    with open(record, "a") as frec:
        frec.write("%s %s %s\n" % (confs, ii, jj))



import glob
import json
import logging
import os
import re

import dpdata
import numpy as np
from monty.serialization import dumpfn, loadfn
from pymatgen.core.structure import Structure
from pymatgen.core.surface import generate_all_slabs

import apex.core.calculator.lib.abacus as abacus
import apex.core.calculator.lib.vasp as vasp
from apex.core.Property import Property
from apex.core.refine import make_refine
from apex.core.reproduce import make_repro, post_repro
from dflow.python import upload_packages
upload_packages.append(__file__)


class Surface(Property):
    def __init__(self, parameter, inter_param=None):
        parameter["reproduce"] = parameter.get("reproduce", False)
        self.reprod = parameter["reproduce"]
        if not self.reprod:
            if not ("init_from_suffix" in parameter and "output_suffix" in parameter):
                self.min_slab_size = parameter["min_slab_size"]
                self.min_vacuum_size = parameter["min_vacuum_size"]
                parameter["pert_xz"] = parameter.get("pert_xz", 0.01)
                self.pert_xz = parameter["pert_xz"]
                default_max_miller = 2
                parameter["max_miller"] = parameter.get(
                    "max_miller", default_max_miller
                )
                self.miller = parameter["max_miller"]
            parameter["cal_type"] = parameter.get("cal_type", "relaxation")
            self.cal_type = parameter["cal_type"]
            default_cal_setting = {
                "relax_pos": True,
                "relax_shape": True,
                "relax_vol": False,
            }
            if "cal_setting" not in parameter:
                parameter["cal_setting"] = default_cal_setting
            else:
                if "relax_pos" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_pos"] = default_cal_setting[
                        "relax_pos"
                    ]
                if "relax_shape" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_shape"] = default_cal_setting[
                        "relax_shape"
                    ]
                if "relax_vol" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_vol"] = default_cal_setting[
                        "relax_vol"
                    ]
            self.cal_setting = parameter["cal_setting"]
        else:
            parameter["cal_type"] = "static"
            self.cal_type = parameter["cal_type"]
            default_cal_setting = {
                "relax_pos": False,
                "relax_shape": False,
                "relax_vol": False,
            }
            if "cal_setting" not in parameter:
                parameter["cal_setting"] = default_cal_setting
            else:
                if "relax_pos" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_pos"] = default_cal_setting[
                        "relax_pos"
                    ]
                if "relax_shape" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_shape"] = default_cal_setting[
                        "relax_shape"
                    ]
                if "relax_vol" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_vol"] = default_cal_setting[
                        "relax_vol"
                    ]
            self.cal_setting = parameter["cal_setting"]
            parameter["init_from_suffix"] = parameter.get("init_from_suffix", "00")
            self.init_from_suffix = parameter["init_from_suffix"]
        self.parameter = parameter
        self.inter_param = inter_param if inter_param != None else {"type": "vasp"}

    def make_confs(self, path_to_work, path_to_equi, refine=False):
        path_to_work = os.path.abspath(path_to_work)
        if os.path.exists(path_to_work):
            #dlog.warning("%s already exists" % path_to_work)
            logging.warning("%s already exists" % path_to_work)
        else:
            os.makedirs(path_to_work)
        path_to_equi = os.path.abspath(path_to_equi)

        if "start_confs_path" in self.parameter and os.path.exists(
            self.parameter["start_confs_path"]
        ):
            init_path_list = glob.glob(
                os.path.join(self.parameter["start_confs_path"], "*")
            )
            struct_init_name_list = []
            for ii in init_path_list:
                struct_init_name_list.append(ii.split("/")[-1])
            struct_output_name = path_to_work.split("/")[-2]
            assert struct_output_name in struct_init_name_list
            path_to_equi = os.path.abspath(
                os.path.join(
                    self.parameter["start_confs_path"],
                    struct_output_name,
                    "relaxation",
                    "relax_task",
                )
            )

        task_list = []
        cwd = os.getcwd()

        if self.reprod:
            print("surface reproduce starts")
            if "init_data_path" not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter["init_data_path"])
            task_list = make_repro(
                self.inter_param,
                init_data_path,
                self.init_from_suffix,
                path_to_work,
                self.parameter.get("reprod_last_frame", True),
            )
            os.chdir(cwd)

        else:
            if refine:
                print("surface refine starts")
                task_list = make_refine(
                    self.parameter["init_from_suffix"],
                    self.parameter["output_suffix"],
                    path_to_work,
                )
                os.chdir(cwd)
                # record miller
                init_from_path = re.sub(
                    self.parameter["output_suffix"][::-1],
                    self.parameter["init_from_suffix"][::-1],
                    path_to_work[::-1],
                    count=1,
                )[::-1]
                task_list_basename = list(map(os.path.basename, task_list))

                for ii in task_list_basename:
                    init_from_task = os.path.join(init_from_path, ii)
                    output_task = os.path.join(path_to_work, ii)
                    os.chdir(output_task)
                    if os.path.isfile("miller.json"):
                        os.remove("miller.json")
                    if os.path.islink("miller.json"):
                        os.remove("miller.json")
                    os.symlink(
                        os.path.relpath(os.path.join(init_from_task, "miller.json")),
                        "miller.json",
                    )
                os.chdir(cwd)

            else:
                if self.inter_param["type"] == "abacus":
                    CONTCAR = abacus.final_stru(path_to_equi)
                    POSCAR = "STRU"
                else:
                    CONTCAR = "CONTCAR"
                    POSCAR = "POSCAR"

                equi_contcar = os.path.join(path_to_equi, CONTCAR)
                if not os.path.exists(equi_contcar):
                    raise RuntimeError("please do relaxation first")

                if self.inter_param["type"] == "abacus":
                    stru = dpdata.System(equi_contcar, fmt="stru")
                    stru.to("contcar", "CONTCAR.tmp")
                    ptypes = vasp.get_poscar_types("CONTCAR.tmp")
                    ss = Structure.from_file("CONTCAR.tmp")
                    os.remove("CONTCAR.tmp")
                else:
                    ptypes = vasp.get_poscar_types(equi_contcar)
                    # gen structure
                    ss = Structure.from_file(equi_contcar)

                # gen slabs
                all_slabs = generate_all_slabs(
                    ss, self.miller, self.min_slab_size, self.min_vacuum_size
                )

                os.chdir(path_to_work)
                if os.path.isfile(POSCAR):
                    os.remove(POSCAR)
                if os.path.islink(POSCAR):
                    os.remove(POSCAR)
                os.symlink(os.path.relpath(equi_contcar), POSCAR)
                #           task_poscar = os.path.join(output, 'POSCAR')
                for ii in range(len(all_slabs)):
                    output_task = os.path.join(path_to_work, "task.%06d" % ii)
                    os.makedirs(output_task, exist_ok=True)
                    os.chdir(output_task)
                    for jj in [
                        "INCAR",
                        "POTCAR",
                        "POSCAR",
                        "conf.lmp",
                        "in.lammps",
                        "STRU",
                    ]:
                        if os.path.exists(jj):
                            os.remove(jj)
                    task_list.append(output_task)
                    print(
                        "# %03d generate " % ii,
                        output_task,
                        " \t %d atoms" % len(all_slabs[ii].sites),
                    )
                    # make confs
                    all_slabs[ii].to("POSCAR.tmp", "POSCAR")
                    vasp.regulate_poscar("POSCAR.tmp", "POSCAR")
                    vasp.sort_poscar("POSCAR", "POSCAR", ptypes)
                    vasp.perturb_xz("POSCAR", "POSCAR", self.pert_xz)
                    if self.inter_param["type"] == "abacus":
                        abacus.poscar2stru("POSCAR", self.inter_param, "STRU")
                        os.remove("POSCAR")
                    # record miller
                    dumpfn(all_slabs[ii].miller_index, "miller.json")
                os.chdir(cwd)

        return task_list

    def post_process(self, task_list):
        pass

    def task_type(self):
        return self.parameter["type"]

    def task_param(self):
        return self.parameter

    def _compute_lower(self, output_file, all_tasks, all_res):
        output_file = os.path.abspath(output_file)
        res_data = {}
        ptr_data = os.path.dirname(output_file) + "\n"

        if not self.reprod:
            ptr_data += "Miller_Indices: \tSurf_E(J/m^2) EpA(eV) equi_EpA(eV)\n"
            for ii in all_tasks:
                task_result = loadfn(os.path.join(ii, "result_task.json"))
                natoms = np.sum(task_result["atom_numbs"])
                epa = task_result["energies"][-1] / natoms
                AA = np.linalg.norm(
                    np.cross(task_result["cells"][0][0], task_result["cells"][0][1])
                )

                equi_path = os.path.abspath(
                    os.path.join(
                        os.path.dirname(output_file), "../relaxation/relax_task"
                    )
                )
                equi_result = loadfn(os.path.join(equi_path, "result.json"))
                equi_epa = equi_result["energies"][-1] / np.sum(
                    equi_result["atom_numbs"]
                )
                structure_dir = os.path.basename(ii)

                Cf = 1.60217657e-16 / (1e-20 * 2) * 0.001
                evac = (task_result["energies"][-1] - equi_epa * natoms) / AA * Cf

                miller_index = loadfn(os.path.join(ii, "miller.json"))
                ptr_data += "%-25s     %7.3f    %8.3f %8.3f\n" % (
                    str(miller_index) + "-" + structure_dir + ":",
                    evac,
                    epa,
                    equi_epa,
                )
                res_data[str(miller_index) + "-" + structure_dir] = [
                    evac,
                    epa,
                    equi_epa,
                ]

        else:
            if "init_data_path" not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter["init_data_path"])
            res_data, ptr_data = post_repro(
                init_data_path,
                self.parameter["init_from_suffix"],
                all_tasks,
                ptr_data,
                self.parameter.get("reprod_last_frame", True),
            )

        with open(output_file, "w") as fp:
            json.dump(res_data, fp, indent=4)

        return res_data, ptr_data



import inspect
from dataclasses import dataclass, field

from dflow import config, s3_config
from dflow.plugins import bohrium
from dflow.plugins.bohrium import TiefblueClient
from dflow.plugins.dispatcher import DispatcherExecutor

from apex.utils import update_dict


@dataclass
class Config:
    # dflow config
    dflow_config: dict = None
    dflow_host: str = "https://127.0.0.1:2746"
    k8s_api_server: str = "https://127.0.0.1:2746"
    mode: str = 'default'
    debug_copy_method: str = "copy"
    debug_pool_workers: int = 1
    debug_workdir: str = '.'

    # dflow s3 config
    dflow_s3_config: dict = None

    # Bohrium config
    bohrium_config: dict = None
    phone: str = None
    email: str = None
    password: str = None
    program_id: int = None

    # DispachterExecutor config
    dispatcher_config: dict = None
    dispatcher_image: str = None
    dispatcher_command: str = "python3"
    dispatcher_remote_command: str = None
    image_pull_policy: str = "IfNotPresent"
    scass_type: str = None
    machine: dict = None
    resources: dict = None
    task: dict = None
    context_type: str = None
    batch_type: str = None
    clean_asynchronously: bool = True
    local_root: str = '.'
    remote_root: str = None
    remote_host: str = None
    remote_username: str = None
    remote_password: str = None
    port: int = 22

    # calculator config
    run_image_name: str = None
    run_command: str = None
    apex_image_name: str = "zhuoyli/apex_amd64"
    group_size: int = None
    pool_size: int = None
    upload_python_packages: list = field(default_factory=list)
    # lammps
    lammps_image_name: str = None
    lammps_run_command: str = None
    # vasp
    vasp_image_name: str = None
    vasp_run_command: str = None
    # abacus
    abacus_image_name: str = None
    abacus_run_command: str = None

    is_bohrium_dflow: bool = False

    def __post_init__(self):
        # judge if running dflow on the Bohrium
        try:
            assert self.dflow_config["host"] == "https://workflows.deepmodeling.com"
        except (AssertionError, TypeError):
            if self.dflow_host == "https://workflows.deepmodeling.com":
                self.is_bohrium_dflow = True
        else:
            self.is_bohrium_dflow = True

        # pre-define machine dictionaries for dpdispatcher
        if not (self.context_type or self.machine):
            self.machine_dict = None
        elif self.context_type in ["Bohrium", "bohrium",
                                    "BohriumContext", "boriumcontext"]:
            self.machine_dict = {
                "batch_type": self.batch_type,
                "context_type": self.context_type,
                "remote_profile": {
                    "email": self.email,
                    "password": self.password,
                    "program_id": self.program_id,
                    "input_data": {
                        "job_type": "container",
                        "platform": "ali",
                        "scass_type": self.scass_type,
                    },
                },
            }
            if self.machine:
                update_dict(self.machine_dict, self.machine)
        elif self.context_type in ["SSHContext", "sshcontext",
                                    "SSH", "ssh"]:
            self.machine_dict = {
                "batch_type": self.batch_type,
                "context_type": self.context_type,
                "local_root": self.local_root,
                "remote_root": self.remote_root,
                "clean_asynchronously": self.clean_asynchronously,
                "remote_profile": {
                    "hostname": self.remote_host,
                    "username": self.remote_username,
                    "password": self.remote_password,
                    "port": self.port,
                    "timeout": 10
                }
            }
            if self.machine:
                update_dict(self.machine_dict, self.machine)
        elif self.context_type in ["LocalContext", "localcontext"
                                    "Local", "local"]:
            self.machine_dict = {
                "batch_type": self.batch_type,
                "context_type": self.context_type,
                "local_root": self.local_root,
                "remote_root": self.remote_root,
                "clean_asynchronously": self.clean_asynchronously
            }
            if self.machine:
                update_dict(self.machine_dict, self.machine)
        elif self.context_type in ["LazyLocalContext", "lazylocalcontext"
                                    "LazyLocal", "lazylocal"]:
            self.machine_dict = {
                "batch_type": self.batch_type,
                "context_type": self.context_type,
                "local_root": self.local_root,
                "clean_asynchronously": self.clean_asynchronously
            }
            if self.machine:
                update_dict(self.machine_dict, self.machine)
        elif self.machine:
            self.machine_dict = self.machine
        else:
            raise RuntimeError(
                f'please provide a nested machine dictionary '
                f'for context type: {self.context_type}'
            )

        # default pool_size as two size must be indicated simultaneously
        if self.group_size and not self.pool_size:
            self.pool_size = 1

    @property
    def dflow_config_dict(self):
        dflow_config = {
            "host": self.dflow_host,
            "k8s_api_server": self.k8s_api_server,
            "mode": self.mode,
            "debug_copy_method": self.debug_copy_method,
            "debug_pool_workers": self.debug_pool_workers,
            "debug_workdir": self.debug_workdir
        }
        if self.dflow_host:
            update_dict(dflow_config, self.dflow_config)
        return dflow_config

    @property
    def dflow_s3_config_dict(self):
        dflow_s3_config = {}
        if self.is_bohrium_dflow:
            dflow_s3_config = {
                "repo_key": "oss-bohrium",
                "storage_client": TiefblueClient()
            }
        if self.dflow_s3_config:
            update_dict(dflow_s3_config, self.dflow_s3_config)
        return dflow_s3_config

    @property
    def bohrium_config_dict(self):
        bohrium_config = {
            "username": self.email,
            "phone": self.phone,
            "password": self.password,
            "program_id": self.program_id
        }
        if self.bohrium_config:
            update_dict(bohrium_config, self.bohrium_config)
        return bohrium_config

    @property
    def dispatcher_config_dict(self):
        dispatcher_config = {
            "image": self.dispatcher_image,
            "image_pull_policy": self.image_pull_policy,
            "machine_dict": self.machine_dict,
            "resources_dict": self.resources,
            "task_dict": self.task,
            "command": self.dispatcher_command,
            "remote_command": self.dispatcher_remote_command
        }
        if self.dispatcher_config:
            update_dict(dispatcher_config, self.dispatcher_config)
        return dispatcher_config

    @property
    def basic_config_dict(self):
        basic_config = {
            "run_image_name": self.run_image_name,
            "run_command": self.run_command,
            "apex_image_name": self.apex_image_name,
            "group_size": self.group_size,
            "pool_size": self.pool_size,
            "upload_python_packages": self.upload_python_packages,
            "lammps_image_name": self.lammps_image_name,
            "lammps_run_command": self.lammps_run_command,
            "vasp_image_name": self.vasp_image_name,
            "vasp_run_command": self.vasp_run_command,
            "abacus_image_name": self.abacus_image_name,
            "abacus_run_command": self.abacus_run_command
        }
        return basic_config

    @staticmethod
    def config_dflow(dflow_config_data: dict) -> None:
        dflow_s3_config = {}
        for kk in dflow_config_data.keys():
            if kk[:3] == "s3_":
                dflow_s3_config[kk[3:]] = dflow_config_data.pop(kk)
        for kk in dflow_config_data.keys():
            config[kk] = dflow_config_data[kk]
        for kk in dflow_s3_config.keys():
            s3_config[kk] = dflow_s3_config[kk]

    @staticmethod
    def config_s3(dflow_s3_config_data: dict) -> None:
        for kk in dflow_s3_config_data.keys():
            s3_config[kk] = dflow_s3_config_data[kk]

    @staticmethod
    def config_bohrium(bohrium_config_data: dict) -> None:
        for kk in bohrium_config_data.keys():
            bohrium.config[kk] = bohrium_config_data[kk]

    def get_executor(
            self,
            dispatcher_config: dict
    ) -> DispatcherExecutor:
        if not (self.context_type or self.machine):
            executor = None
        else:
            # get arguments for instantiation of the DispatcherExecutor
            sig = inspect.signature(DispatcherExecutor.__init__)
            # pop out 'self'
            defined_parameters = list(sig.parameters.keys())
            defined_parameters.pop(0)
            # filter the dispatcher_config
            filtered_parameters = {
                k: v for k, v in dispatcher_config.items() if k in defined_parameters
            }
            executor = DispatcherExecutor(**filtered_parameters)

        return executor



import os
from pathlib import (
    Path,
)
from typing import (
    List,
    Optional,
    Type,
)
from dflow import (
    InputArtifact,
    InputParameter,
    Inputs,
    OutputArtifact,
    Outputs,
    Step,
    Steps,
    argo_range,
    argo_enumerate
)
from dflow.python import (
    OP,
    PythonOPTemplate,
    Slices,
)
from dflow.plugins.dispatcher import DispatcherExecutor
from apex.op.property_ops import DistributeProps, CollectProps
from apex.superop.SimplePropertySteps import SimplePropertySteps


class PropertyFlow(Steps):

    def __init__(
        self,
        name: str,
        make_op: Type[OP],
        run_op: Type[OP],
        post_op: Type[OP],
        make_image: str,
        run_image: str,
        post_image: str,
        run_command: str,
        calculator: str,
        group_size: Optional[int] = None,
        pool_size: Optional[int] = None,
        executor: Optional[DispatcherExecutor] = None,
        upload_python_packages: Optional[List[os.PathLike]] = None,
    ):
        self._input_parameters = {
            "flow_id": InputParameter(type=str, value=""),
            "parameter": InputParameter(type=dict)
        }
        self._input_artifacts = {
            "input_work_path": InputArtifact(type=Path)
        }
        self._output_parameters = {

        }
        self._output_artifacts = {
            "retrieve_path": OutputArtifact(type=List[Path])
        }

        super().__init__(
            name=name,
            inputs=Inputs(
                parameters=self._input_parameters,
                artifacts=self._input_artifacts
            ),
            outputs=Outputs(
                parameters=self._output_parameters,
                artifacts=self._output_artifacts
            ),
        )

        self._keys = ["distributor", "propcal", "collector"]
        self.step_keys = {}
        key = "distributor"
        self.step_keys[key] = '--'.join(
            [str(self.inputs.parameters["flow_id"]), key]
        )
        key = "propcal"
        self.step_keys[key] = '--'.join(
            [str(self.inputs.parameters["flow_id"]), key]
        )
        key = "collector"
        self.step_keys[key] = '--'.join(
            [str(self.inputs.parameters["flow_id"]), key]
        )

        self._build(
            name,
            make_op,
            run_op,
            post_op,
            make_image,
            run_image,
            post_image,
            run_command,
            calculator,
            group_size,
            pool_size,
            executor,
            upload_python_packages
        )

    @property
    def input_parameters(self):
        return self._input_parameters

    @property
    def input_artifacts(self):
        return self._input_artifacts

    @property
    def output_parameters(self):
        return self._output_parameters

    @property
    def output_artifacts(self):
        return self._output_artifacts

    @property
    def keys(self):
        return self._keys

    def _build(
        self,
        name: str,
        make_op: Type[OP],
        run_op: Type[OP],
        post_op: Type[OP],
        make_image: str,
        run_image: str,
        post_image: str,
        run_command: str,
        calculator: str,
        group_size: Optional[int] = None,
        pool_size: Optional[int] = None,
        executor: Optional[DispatcherExecutor] = None,
        upload_python_packages: Optional[List[os.PathLike]] = None,
    ):
        distributor = Step(
            name="Distributor",
            template=PythonOPTemplate(DistributeProps,
                                      image=make_image,
                                      command=["python3"]),
            artifacts={"input_work_path": self.inputs.artifacts["input_work_path"]},
            parameters={"param": self.inputs.parameters["parameter"]},
            key="distributor"
        )
        self.add(distributor)

        simple_property_steps = SimplePropertySteps(
            name='simple-property-flow',
            make_op=make_op,
            run_op=run_op,
            post_op=post_op,
            make_image=make_image,
            run_image=run_image,
            post_image=post_image,
            run_command=run_command,
            calculator=calculator,
            group_size=group_size,
            pool_size=pool_size,
            executor=executor,
            upload_python_packages=upload_python_packages
        )

        propscal = Step(
            name="Prop-Cal",
            template=simple_property_steps,
            slices=Slices(
                slices="{{item.order}}",
                input_parameter=[
                    "path_to_prop",
                    "prop_param",
                    "inter_param",
                    "do_refine"
                ],
                input_artifact=["input_work_path"],
                output_artifact=["output_post"],
            ),
            artifacts={
                "input_work_path": distributor.outputs.artifacts["orig_work_path"]
            },
            parameters={
                "flow_id": "{{item.value}}",
                "path_to_prop": distributor.outputs.parameters["path_to_prop"],
                "prop_param": distributor.outputs.parameters["prop_param"],
                "inter_param": distributor.outputs.parameters["inter_param"],
                "do_refine": distributor.outputs.parameters["do_refine"]
            },
            with_param=argo_enumerate(distributor.outputs.parameters["flow_id"]),
            key="propscal-{{item}}"
        )
        self.add(propscal)

        collector = Step(
            name="PropsCollector",
            template=PythonOPTemplate(CollectProps,
                                      image=make_image,
                                      command=["python3"]),
            artifacts={"input_all": self.inputs.artifacts["input_work_path"],
                       "input_post": propscal.outputs.artifacts["output_post"]},
            parameters={"param": self.inputs.parameters["parameter"]},
            key="collector"
        )
        self.add(collector)

        self.outputs.artifacts["retrieve_path"]._from \
            = collector.outputs.artifacts["retrieve_path"]



import numpy as np
from dpdata.periodic_table import Element
from dflow.python import upload_packages
upload_packages.append(__file__)


def _make_siesta_01_common(sys_data, ecut, ediff, mixingWeight, NumberPulay):
    tot_natoms = sum(sys_data["atom_numbs"])
    ntypes = len(sys_data["atom_names"])
    ret = ""
    ret += "SystemName        system\n"
    ret += "SystemLabel       system\n"
    ret += "NumberOfAtoms     %d" % tot_natoms
    ret += "\nNumberOfSpecies   %d\n" % ntypes
    ret += "\n"
    ret += "WriteForces       T\n"
    ret += "WriteCoorStep     T\n"
    ret += "WriteCoorXmol     T\n"
    ret += "WriteMDXmol       T\n"
    ret += "WriteMDHistory    T\n\n"

    ret += "MeshCutoff            %s" % str(ecut)
    ret += " Ry\n"
    ret += "DM.MixingWeight       %f\n" % mixingWeight
    ret += "DM.Tolerance          %e\n" % ediff
    ret += "DM.UseSaveDM          true\n"
    ret += "DM.NumberPulay         %d\n" % NumberPulay
    ret += "MD.UseSaveXV           T\n\n"

    ret += "XC.functional          GGA\n"
    ret += "XC.authors             PBE\n"
    ret += "DM.UseSaveDM           F\n"
    ret += "WriteDM                F\n"
    ret += "WriteDM.NetCDF         F\n"
    ret += "WriteDMHS.NetCDF       F\n"
    return ret


def _make_siesta_02_species(sys_data, pps):
    atom_nums = sys_data["atom_numbs"]
    atom_names = sys_data["atom_names"]
    ntypes = len(atom_nums)
    assert ntypes == len(atom_names)
    assert ntypes == len(pps)
    ret = ""
    ret += "%block Chemical_Species_label\n"
    for i in range(0, len(atom_names)):
        ret += (
            str(i + 1)
            + "\t"
            + str(Element(atom_names[i]).Z)
            + "\t"
            + atom_names[i]
            + "\n"
        )
    ret += "%endblock Chemical_Species_label\n"
    return ret


# ## kpoints !!! can not understand
def _make_siesta_03_kpoint(sys_data, kspacing):
    cell = sys_data["cells"][0]
    cell = np.reshape(cell, [3, 3])
    ## np.linalg.inv()：矩阵求逆
    rcell = np.linalg.inv(cell)
    ## .T 矩阵转置
    rcell = rcell.T
    # np.ceil()是向上取整，与四舍五入无关 -5.6 --> -5
    # np.linalg.norm：进行范数运算，范数是对向量（或者矩阵）的度量，是一个标量（scalar）
    kpoints = [
        (np.ceil(2 * np.pi * np.linalg.norm(ii) / kspacing).astype(int)) for ii in rcell
    ]
    ret = ""
    ret += "%block kgrid_Monkhorst_Pack\n"
    ret += "%d" % kpoints[0]
    ret += "\t0\t0\t0.0\n"

    ret += "0\t"
    ret += "%d" % kpoints[1]
    ret += "\t0\t0.0\n"

    ret += "0\t0\t"
    ret += "%d" % kpoints[2]
    ret += "\t0.0\n"

    ret += "%endblock kgrid_Monkhorst_Pack\n"
    return ret


### coordinate
def _make_siesta_04_ucVectorCoord(sys_data):
    cell = sys_data["cells"][0]
    cell = np.reshape(cell, [3, 3])
    coordinates = sys_data["coords"][0]
    atom_names = sys_data["atom_names"]
    atom_numbs = sys_data["atom_numbs"]
    ntypes = len(atom_names)
    ret = ""
    ret += "LatticeConstant try_input_output.00 Ang\n"
    ret += "%block LatticeVectors\n"
    for ii in range(3):
        for jj in range(3):
            ret += "%f " % cell[ii][jj]
        ret += "\n"
    ret += "%endblock LatticeVectors\n"

    ret += "\n"
    ret += "AtomicCoordinatesFormat Ang\n"
    ret += "%block AtomicCoordinatesAndAtomicSpecies\n"
    cc = 0
    for ii in range(ntypes):
        for jj in range(atom_numbs[ii]):
            ret += "%f %f %f %d %s\n" % (
                coordinates[cc][0],
                coordinates[cc][1],
                coordinates[cc][2],
                ii + 1,
                atom_names[ii],
            )
            cc += 1
    ret += "%endblock AtomicCoordinatesAndAtomicSpecies"
    return ret


def make_siesta_input(sys_data, fp_pp_files, fp_params):
    ecut = fp_params["ecut"]
    ediff = fp_params["ediff"]
    mixingWeight = fp_params["mixingWeight"]
    NumberPulay = fp_params["NumberPulay"]
    kspacing = fp_params["kspacing"]
    ret = ""
    ret += _make_siesta_01_common(sys_data, ecut, ediff, mixingWeight, NumberPulay)
    ret += "\n"
    ret += _make_siesta_02_species(sys_data, fp_pp_files)
    ret += "\n"
    ret += _make_siesta_03_kpoint(sys_data, kspacing)
    ret += "\n"
    ret += _make_siesta_04_ucVectorCoord(sys_data)
    ret += "\n"
    return ret



#!/usr/bin/env python3

import numpy as np
from dflow.python import upload_packages
upload_packages.append(__file__)


def _get_block(lines, keys):
    for idx in range(len(lines)):
        if keys in lines[idx]:
            break
    if idx == len(lines) - 1:
        return None
    idx_s = idx + 2
    idx = idx_s
    ret = []
    while True:
        if len(lines[idx].split()) == 0:
            break
        else:
            ret.append(lines[idx])
        idx += 1
    return ret


def lmpbox2box(lohi, tilt):
    xy = tilt[0]
    xz = tilt[1]
    yz = tilt[2]
    orig = np.array([lohi[0][0], lohi[1][0], lohi[2][0]])
    lens = []
    for dd in range(3):
        lens.append(lohi[dd][1] - lohi[dd][0])
    xx = [lens[0], 0, 0]
    yy = [xy, lens[1], 0]
    zz = [xz, yz, lens[2]]
    return orig, np.array([xx, yy, zz])


def box2lmpbox(orig, box):
    lohi = np.zeros([3, 2])
    for dd in range(3):
        lohi[dd][0] = orig[dd]
    tilt = np.zeros(3)
    tilt[0] = box[1][0]
    tilt[1] = box[2][0]
    tilt[2] = box[2][1]
    lens = np.zeros(3)
    lens[0] = box[0][0]
    lens[1] = box[1][1]
    lens[2] = box[2][2]
    for dd in range(3):
        lohi[dd][1] = lohi[dd][0] + lens[dd]
    return lohi, tilt


def get_atoms(lines):
    return _get_block(lines, "Atoms")


def get_natoms(lines):
    for ii in lines:
        if "atoms" in ii:
            return int(ii.split()[0])
    return None


def get_natomtypes(lines):
    for ii in lines:
        if "atom types" in ii:
            return int(ii.split()[0])
    return None


def _atom_info_mol(line):
    vec = line.split()
    # idx, mole_type, atom_type, charge, x, y, z
    return (
        int(vec[0]),
        int(vec[1]),
        int(vec[2]),
        float(vec[3]),
        float(vec[4]),
        float(vec[5]),
        float(vec[6]),
    )


def _atom_info_atom(line):
    vec = line.split()
    # idx, atom_type, x, y, z
    return int(vec[0]), int(vec[1]), float(vec[2]), float(vec[3]), float(vec[4])


def get_natoms_vec(lines):
    atype = get_atype(lines)
    natoms_vec = []
    natomtypes = get_natomtypes(lines)
    for ii in range(natomtypes):
        natoms_vec.append(sum(atype == ii + 1))
    assert sum(natoms_vec) == get_natoms(lines)
    return natoms_vec


def get_atype(lines):
    alines = get_atoms(lines)
    atype = []
    for ii in alines:
        # idx, mt, at, q, x, y, z = _atom_info_mol(ii)
        idx, at, x, y, z = _atom_info_atom(ii)
        atype.append(at)
    return np.array(atype, dtype=int)


def get_posi(lines):
    atom_lines = get_atoms(lines)
    posis = []
    for ii in atom_lines:
        # posis.append([float(jj) for jj in ii.split()[4:7]])
        posis.append([float(jj) for jj in ii.split()[2:5]])
    return np.array(posis)


def get_lmpbox(lines):
    box_info = []
    tilt = np.zeros(3)
    for ii in lines:
        if "xlo" in ii and "xhi" in ii:
            box_info.append([float(ii.split()[0]), float(ii.split()[1])])
            break
    for ii in lines:
        if "ylo" in ii and "yhi" in ii:
            box_info.append([float(ii.split()[0]), float(ii.split()[1])])
            break
    for ii in lines:
        if "zlo" in ii and "zhi" in ii:
            box_info.append([float(ii.split()[0]), float(ii.split()[1])])
            break
    for ii in lines:
        if "xy" in ii and "xz" in ii and "yz" in ii:
            tilt = np.array([float(jj) for jj in ii.split()[0:3]])
    return box_info, tilt


def system_data(lines):
    system = {}
    system["atom_numbs"] = get_natoms_vec(lines)
    system["atom_names"] = []
    for ii in range(len(system["atom_numbs"])):
        system["atom_names"].append("Type_%d" % ii)
    lohi, tilt = get_lmpbox(lines)
    orig, cell = lmpbox2box(lohi, tilt)
    system["orig"] = np.array(orig)
    system["cell"] = np.array(cell)
    natoms = sum(system["atom_numbs"])
    system["atom_types"] = get_atype(lines)
    system["coordinates"] = get_posi(lines)
    return system


def to_system_data(lines):
    return system_data(lines)


def from_system_data(system):
    ret = ""
    ret += "\n"
    natoms = sum(system["atom_numbs"])
    ntypes = len(system["atom_numbs"])
    ret += "%d atoms\n" % natoms
    ret += "%d atom types\n" % ntypes
    ret += "0 %f xlo xhi\n" % system["cell"][0][0]
    ret += "0 %f ylo yhi\n" % system["cell"][1][1]
    ret += "0 %f zlo zhi\n" % system["cell"][2][2]
    ret += "%f %f %f xy xz yz\n" % (
        system["cell"][1][0],
        system["cell"][2][0],
        system["cell"][2][1],
    )
    ret += "\n"
    ret += "Atoms # atomic\n"
    ret += "\n"
    for ii in range(natoms):
        ret += "%d %d %f %f %f\n" % (
            ii + 1,
            system["atom_types"][ii],
            system["coordinates"][ii][0] - system["orig"][0],
            system["coordinates"][ii][1] - system["orig"][1],
            system["coordinates"][ii][2] - system["orig"][2],
        )
    return ret


if __name__ == "__main__":
    fname = "water-SPCE.data"
    lines = open(fname).read().split("\n")
    bonds, tilt = get_lmpbox(lines)
    # print(bonds, tilt)
    orig, box = lmpbox2box(bonds, tilt)
    # print(orig, box)
    bonds1, tilt1 = box2lmpbox(orig, box)
    # print(bonds1, tilt1)
    print(bonds1 - bonds)
    print(tilt1 - tilt)
    print(box)
    print(get_atype(lines))
    print(get_posi(lines))



import glob
import json
import os
import shutil
import sys
import unittest

import dpdata
import numpy as np
from monty.serialization import dumpfn, loadfn

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"

from pymatgen.io.vasp import Incar

from apex.core.common_prop import make_property

from .context import make_kspacing_kpoints, setUpModule


class TestMakeProperty(unittest.TestCase):
    jdata = {
        "structures": ["confs/std-fcc"],
        "interaction": {
            "type": "vasp",
            "incar": "vasp_input/INCAR.rlx",
            "potcar_prefix": "vasp_input",
            "potcars": {"Al": "POT_Al"},
        },
        "properties": [
            {
                "type": "eos",
                "skip": False,
                "vol_start": 0.8,
                "vol_end": 1.2,
                "vol_step": 0.01,
                "cal_setting": {
                    "relax_pos": True,
                    "relax_shape": True,
                    "relax_vol": False,
                    "overwrite_interaction": {
                        "type": "vasp",
                        "incar": "vasp_input/INCAR.rlx",
                        "potcar_prefix": "vasp_input",
                        "potcars": {"Al": "POT_Al"},
                    },
                },
            }
        ],
    }

    def tearDown(self):
        if os.path.exists("confs/std-fcc/eos_00"):
            shutil.rmtree("confs/std-fcc/eos_00")
        if os.path.exists("confs/std-fcc/relaxation"):
            shutil.rmtree("confs/std-fcc/relaxation")

    def test_make_eos(self):
        confs = self.jdata["structures"]
        inter_param = self.jdata["interaction"]
        property_list = self.jdata["properties"]

        target_path = "confs/std-fcc/eos_00"
        equi_path = "confs/std-fcc/relaxation/relax_task"
        source_path = "equi/vasp"

        if not os.path.exists(equi_path):
            os.makedirs(equi_path)
        shutil.copy(
            os.path.join(source_path, "CONTCAR_Al_fcc"),
            os.path.join(equi_path, "CONTCAR"),
        )

        make_property(confs, inter_param, property_list)

        dfm_dirs = glob.glob(os.path.join(target_path, "task.*"))

        incar0 = Incar.from_file(os.path.join("vasp_input", "INCAR.rlx"))
        incar0["ISIF"] = 4

        with open(os.path.join("vasp_input", "POT_Al")) as fp:
            pot0 = fp.read()
        for ii in dfm_dirs:
            self.assertTrue(os.path.isfile(os.path.join(ii, "KPOINTS")))
            incar1 = Incar.from_file(os.path.join(ii, "INCAR"))
            self.assertTrue(incar0 == incar1)
            self.assertTrue(os.path.isfile(os.path.join(ii, "INCAR")))
            self.assertTrue(os.path.isfile(os.path.join(ii, "POSCAR")))
            self.assertTrue(os.path.isfile(os.path.join(ii, "POTCAR")))
            self.assertTrue(os.path.isfile(os.path.join(ii, "task.json")))
            inter_json_file = os.path.join(ii, "inter.json")
            self.assertTrue(os.path.isfile(inter_json_file))
            inter_json = loadfn(inter_json_file)
            self.assertEqual(inter_json, inter_param)
            self.assertEqual(
                os.path.realpath(os.path.join(ii, "POSCAR.orig")),
                os.path.realpath(os.path.join(equi_path, "CONTCAR")),
            )
            with open(os.path.join(ii, "POTCAR")) as fp:
                poti = fp.read()
            self.assertEqual(pot0, poti)



import glob
import os
import shutil
import sys
import unittest

import numpy as np
from pymatgen.analysis.defects.core import Vacancy as pmg_Vacancy
from pymatgen.core import Structure
from pymatgen.io.vasp import Incar
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from apex.core.property.Vacancy import Vacancy

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"


class TestVacancy(unittest.TestCase):
    def setUp(self):
        _jdata = {
            "structures": ["confs/hp-Li"],
            "interaction": {
                "type": "vasp",
                "incar": "vasp_input/INCAR.rlx",
                "potcar_prefix": ".",
                "potcars": {"Yb": "vasp_input/POTCAR"},
            },
            "properties": [{"type": "vacancy", "supercell": [1, 1, 1]}],
        }

        self.equi_path = "confs/hp-Li/relaxation/relax_task"
        self.source_path = "equi/vasp"
        self.target_path = "confs/hp-Li/vacancy_00"
        if not os.path.exists(self.equi_path):
            os.makedirs(self.equi_path)

        self.confs = _jdata["structures"]
        self.inter_param = _jdata["interaction"]
        self.prop_param = _jdata["properties"]

        self.vacancy = Vacancy(_jdata["properties"][0])

    def tearDown(self):
        if os.path.exists(self.equi_path):
            shutil.rmtree(self.equi_path)
        if os.path.exists(self.target_path):
            shutil.rmtree(self.target_path)

    def test_task_type(self):
        self.assertEqual("vacancy", self.vacancy.task_type())

    def test_task_param(self):
        self.assertEqual(self.prop_param[0], self.vacancy.task_param())

    def test_make_confs_0(self):
        if not os.path.exists(os.path.join(self.equi_path, "CONTCAR")):
            with self.assertRaises(RuntimeError):
                self.vacancy.make_confs(self.target_path, self.equi_path)
        shutil.copy(
            os.path.join(self.source_path, "CONTCAR"),
            os.path.join(self.equi_path, "CONTCAR"),
        )
        task_list = self.vacancy.make_confs(self.target_path, self.equi_path)
        dfm_dirs = glob.glob(os.path.join(self.target_path, "task.*"))
        self.assertEqual(len(dfm_dirs), 5)

        incar0 = Incar.from_file(os.path.join("vasp_input", "INCAR.rlx"))
        incar0["ISIF"] = 4

        self.assertEqual(
            os.path.realpath(os.path.join(self.equi_path, "CONTCAR")),
            os.path.realpath(os.path.join(self.target_path, "POSCAR")),
        )
        ref_st = Structure.from_file(os.path.join(self.target_path, "POSCAR"))
        sga = SpacegroupAnalyzer(ref_st)
        sym_st = sga.get_symmetrized_structure()
        equiv_site_seq = list(sym_st.equivalent_sites)
        dfm_dirs.sort()
        for ii in dfm_dirs:
            st_file = os.path.join(ii, "POSCAR")
            self.assertTrue(os.path.isfile(st_file))
            st0 = Structure.from_file(st_file)
            vac_site = equiv_site_seq.pop(0)
            vac = pmg_Vacancy(ref_st, vac_site[0])
            st1 = vac.get_supercell_structure(
                sc_mat=np.eye(3) * self.prop_param[0]["supercell"]
            )
        for ii in range(3):
            st0_coords = [format(st0.sites[ii].coords[jj], '.4f') for jj in range(3)]
            st1_coords = [format(st1.sites[ii].coords[jj], '.4f') for jj in range(3)]
        self.assertEqual(st0_coords, st1_coords)
